# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities to create procedurally generated matrices."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np


class PGM(object):
  """PGM where ground-truh factors are represented as integer values."""

  def __init__(self,
               random_state,
               num_relations,
               atom_counts,
               sampling_strategy="easy",
               num_rows=3,
               num_cols=3,
               num_solutions=6):
    """Creates a PGM.

    Args:
      random_state: np.random.RandomState used to sample the PGM.
      num_relations: Number of relations to enforce for each row in the PGM.
      atom_counts: List that contains the number of atoms for each of the
        ground-truth factors.
      sampling_strategy: Either `easy` or `hard`. For `easy`, alternative
        answers are random other solutions that do not satisfy the constraints
        in the given PGM. For `hard`, alternative answers are unique random
        modifications of the correct solution which makes the task  harder.
      num_rows: Integer with number of rows.
      num_cols: Integer with number of columns.
      num_solutions: Integer with number of solutions in the PGM.
    """
    if sampling_strategy == "easy":
      sampling_fn = sample_easy_alternative
    elif sampling_strategy == "hard":
      sampling_fn = sample_hard_alternative
    else:
      raise ValueError("Only easy and hard sampling are currently supported.")

    # Create a factory for this PGM and sample a random solution to it.
    self.design = PGMDesign(random_state, num_relations, atom_counts, num_rows,
                            num_cols)
    self.matrix = self.design.sample()

    # Use the sample strategy to create additional wrong solutions.
    self.other_solutions = []
    for _ in range(num_solutions - 1):
      self.other_solutions.append(
          sampling_fn(self.design, self.matrix, self.other_solutions))
    self.other_solutions = np.array(self.other_solutions)

  def print_pgm(self):
    """Prints the PGM to stdout."""
    for i in range(self.matrix.shape[2]):
      print("---")
      print("Factor %d" % i)
      print("---")
      print("Solution:")
      print(self.matrix[:, :, i])
      print("Alternatives:")
      print(self.other_solutions[:, i])


class PGMDesign(object):
  """Captures the design of a PGM (i.e. the rules) but not the actual values."""

  def __init__(self,
               random_state,
               num_relations,
               atom_counts,
               num_rows=3,
               num_cols=3):
    """Creates a PGMDesign.

    Args:
      random_state: np.random.RandomState used to sample the PGM.
      num_relations: Number of relations to enforce for each row in the PGM.
      atom_counts: List that contains the number of atoms for each of the
        ground-truth factors.
      num_rows: Integer with number of rows.
      num_cols: Integer with number of columns.
    """
    self.random_state = random_state
    self.num_relations = num_relations
    self.atom_counts = atom_counts
    self.num_rows = num_rows
    self.num_cols = num_cols

    # Setup list to keep the relations for each of the factors. By default,
    # each factor has no active relation.
    self.relations = [NonActiveRelation for _ in atom_counts]

    # Randomly sample factors where all factors will be the same across rows.
    self.num_factors = len(atom_counts)
    if self.num_factors < num_relations:
      raise ValueError("Cannot have less factors than relations.")
    indices = list(range(self.num_factors))
    self.active_relations = []
    for _ in range(num_relations):
      selected_index = indices.pop(random_state.choice(len(indices)))
      self.active_relations.append(selected_index)
      self.relations[selected_index] = ConstantRelation

    # Create the actual relations.
    for i, num_atoms in enumerate(atom_counts):
      self.relations[i] = self.relations[i](num_atoms, num_rows, num_cols)

  def sample(self):
    """Sample the actual values of a PGM.

    Returns:
      Numpy array of type np.int64 and shape (num_rows, num_cols, num_factors)
        with the ground-truth factor values of the PGM.
    """
    matrix = np.zeros((self.num_rows, self.num_cols, self.num_factors),
                      dtype=np.int64)
    for i, relation in enumerate(self.relations):
      matrix[:, :, i] = relation.sample(self.random_state)
    return matrix

  def randomly_modify_solution(self, initial_solution):
    """Randomly modifies solution to generate hard alternatives.

    Args:
      initial_solution: Numpy array with shape (num_rows, num_cols, num_factors)
        with the ground-truth factor values of the original PGM.

    Returns:
      Numpy array of type np.int64 and shape (num_rows, num_cols, num_factors)
        with the ground-truth factor values of a randomly modified PGM.
    """
    solution = np.copy(initial_solution)
    # Resample a random factor uniformly where a relation is active.
    i = self.random_state.choice(self.active_relations)
    relation = self.relations[i]
    solution[i] = self.random_state.choice(relation.num_atoms)
    # Change all the non-active relations to random values.
    for i, relation in enumerate(self.relations):
      if i not in self.active_relations:
        solution[i] = self.random_state.choice(relation.num_atoms)
    return solution

  def is_consistent(self, matrix):
    """Check whether the matrix is consistent with the PGM Design."""
    for i, relation in enumerate(self.relations):
      if not relation.is_consistent(matrix[:, :, i]):
        return False
    return True

  def resample_design(self):
    """Generates an alternative design of the PGM.

    This will generate a different set of active/non-active relations.

    Returns:
      PGMDesign instance with the new design.
    """
    return PGMDesign(self.random_state, self.num_relations, self.atom_counts,
                     self.num_rows, self.num_cols)


def sample_easy_alternative(design, matrix, already_sampled_alternatives):
  """Samples easy alternative based on sampling a new PGM."""
  for _ in range(100):
    alternative_pgm = design.resample_design().sample()
    # Combine the solutions.
    alternative_solution = np.copy(matrix)
    alternative_solution[-1, -1, :] = alternative_pgm[-1, -1, :]
    if design.is_consistent(alternative_solution):
      continue
    for already_sampled_alternative in already_sampled_alternatives:
      if np.allclose(already_sampled_alternative, alternative_pgm[-1, -1, :]):
        continue
    return alternative_pgm[-1, -1, :]
  raise ValueError("Could not sample alternative solutions.")


def sample_hard_alternative(design, matrix, already_sampled_alternatives):
  """Samples hard alternative based on sampling a new PGM."""
  solution_so_far = matrix[-1, -1]
  for _ in range(100):
    solution_so_far = design.randomly_modify_solution(solution_so_far)
    # Combine the solutions.
    alternative_solution = np.copy(matrix)
    alternative_solution[-1, -1, :] = solution_so_far
    if design.is_consistent(alternative_solution):
      continue
    for already_sampled_alternative in already_sampled_alternatives:
      if np.allclose(already_sampled_alternative, solution_so_far):
        continue
    return solution_so_far
  raise ValueError("Could not sample hard alternative solutions.")


class Relation(object):
  """Abstract base class for relations."""

  def __init__(self, num_atoms, num_rows=3, num_cols=3):
    if num_atoms < num_cols:
      raise ValueError("Cannot have less atoms than columns.")
    if num_atoms == 1:
      raise ValueError("Need more than one atom.")
    self.num_atoms = num_atoms
    self.num_rows = num_rows
    self.num_cols = num_cols

  @staticmethod
  def is_consistent(matrix):
    """Checks whether the matrix satisfies the relation."""
    raise NotImplementedError()

  def sample(self, random_state):
    """Samples a matrix consistent with the relation."""
    raise NotImplementedError()


def is_constant_row(row):
  return len(np.unique(row)) == 1


class ConstantRelation(Relation):
  """Relation where rows in the matrix are constant."""

  @staticmethod
  def is_consistent(matrix):
    """Checks whether the matrix satisfies the relation."""
    for row in matrix:
      if not is_constant_row(row):
        return False
    return True

  def sample(self, random_state):
    """Samples a matrix consistent with the relation."""
    rows = []
    for _ in range(self.num_rows):
      sampled_atom = random_state.choice(self.num_atoms)
      rows.append([sampled_atom] * self.num_cols)
    return np.array(rows)


def is_distinct_row(row):
  return len(np.unique(row)) == len(row)


class DistinctRelation(Relation):
  """Relation where elements in a matrix row are distinct."""

  @staticmethod
  def is_consistent(matrix):
    """Checks whether the matrix satisfies the relation."""
    for row in matrix:
      if not is_distinct_row(row):
        return False
    return True

  def sample(self, random_state):
    """Samples a matrix consistent with the relation."""
    rows = []
    for _ in range(self.num_rows):
      random_permutation = random_state.permutation(self.num_atoms)
      rows.append(random_permutation[:self.num_cols])
    return np.array(rows)


class NonActiveRelation(Relation):
  """Relation where elements are random but do not satisfy other relation."""

  @staticmethod
  def is_consistent(matrix):
    """Checks whether the matrix satisfies the relation."""
    # We need to make sure there are no consistent relations in the rows except
    # the last ones.
    relevant_matrix = matrix[:-1, :]
    for relation in [ConstantRelation, DistinctRelation]:
      if relation.is_consistent(relevant_matrix):
        return False
    return True

  def _sample(self, random_state):
    """Sample a random matrix."""
    return random_state.choice(
        self.num_atoms, size=(self.num_rows, self.num_cols))

  def sample(self, random_state):
    """Samples a matrix consistent with the relation."""
    for _ in range(1000):
      matrix = self._sample(random_state)
      if self.is_consistent(matrix):
        return matrix
    raise ValueError("Could not sample non-relational matrix.")



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Cars3D data set."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import os
from disentanglement_lib.data.ground_truth import ground_truth_data
from disentanglement_lib.data.ground_truth import util
import numpy as np
import PIL
import scipy.io as sio
from six.moves import range
from sklearn.utils import extmath
from tensorflow.compat.v1 import gfile


CARS3D_PATH = os.path.join(
    os.environ.get("DISENTANGLEMENT_LIB_DATA", "."), "cars")



class Cars3D(ground_truth_data.GroundTruthData):
  """Cars3D data set.

  The data set was first used in the paper "Deep Visual Analogy-Making"
  (https://papers.nips.cc/paper/5845-deep-visual-analogy-making) and can be
  downloaded from http://www.scottreed.info/. The images are rescaled to 64x64.

  The ground-truth factors of variation are:
  0 - elevation (4 different values)
  1 - azimuth (24 different values)
  2 - object type (183 different values)
  """

  def __init__(self):
    self.factor_sizes = [4, 24, 183]
    features = extmath.cartesian(
        [np.array(list(range(i))) for i in self.factor_sizes])
    self.latent_factor_indices = [0, 1, 2]
    self.num_total_factors = features.shape[1]
    self.index = util.StateSpaceAtomIndex(self.factor_sizes, features)
    self.state_space = util.SplitDiscreteStateSpace(self.factor_sizes,
                                                    self.latent_factor_indices)
    self.data_shape = [64, 64, 3]
    self.images = self._load_data()

  @property
  def num_factors(self):
    return self.state_space.num_latent_factors

  @property
  def factors_num_values(self):
    return self.factor_sizes


  @property
  def observation_shape(self):
    return self.data_shape

  def sample_factors(self, num, random_state):
    """Sample a batch of factors Y."""
    return self.state_space.sample_latent_factors(num, random_state)

  def sample_observations_from_factors(self, factors, random_state):
    """Sample a batch of observations X given a batch of factors Y."""
    all_factors = self.state_space.sample_all_factors(factors, random_state)
    indices = self.index.features_to_index(all_factors)
    return self.images[indices].astype(np.float32)

  def _load_data(self):
    dataset = np.zeros((24 * 4 * 183, 64, 64, 3))
    all_files = [x for x in gfile.ListDirectory(CARS3D_PATH) if ".mat" in x]
    for i, filename in enumerate(all_files):
      data_mesh = _load_mesh(filename)
      factor1 = np.array(list(range(4)))
      factor2 = np.array(list(range(24)))
      all_factors = np.transpose([
          np.tile(factor1, len(factor2)),
          np.repeat(factor2, len(factor1)),
          np.tile(i,
                  len(factor1) * len(factor2))
      ])
      indexes = self.index.features_to_index(all_factors)
      dataset[indexes] = data_mesh
    return dataset


def _load_mesh(filename):
  """Parses a single source file and rescales contained images."""
  with gfile.Open(os.path.join(CARS3D_PATH, filename), "rb") as f:
    mesh = np.einsum("abcde->deabc", sio.loadmat(f)["im"])
  flattened_mesh = mesh.reshape((-1,) + mesh.shape[2:])
  rescaled_mesh = np.zeros((flattened_mesh.shape[0], 64, 64, 3))
  for i in range(flattened_mesh.shape[0]):
    pic = PIL.Image.fromarray(flattened_mesh[i, :, :, :])
    pic.thumbnail((64, 64, 3), PIL.Image.ANTIALIAS)
    rescaled_mesh[i, :, :, :] = np.array(pic)
  return rescaled_mesh * 1. / 255



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Allows to convolute TFHub modules."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import tensorflow.compat.v1 as tf
import tensorflow_hub as hub
from tensorflow.contrib import framework as contrib_framework


def convolute_and_save(module_path, signature, export_path, transform_fn,
                       transform_checkpoint_path, new_signature=None):
  """Loads TFHub module, convolutes it with transform_fn and saves it again.

  Args:
    module_path: String with path from which the module is constructed.
    signature: String with name of signature to use for loaded module.
    export_path: String with path where to save the final TFHub module.
    transform_fn: Function that creates the graph to be appended to the loaded
      TFHub module. The function should take as keyword arguments the tensors
      returned by the loaded TFHub module. The function should return a
      dictionary of tensor that will be the output of the new TFHub module.
    transform_checkpoint_path: Path to checkpoint from which the transformer_fn
      variables will be read.
    new_signature: String with new name of signature to use for saved module. If
      None, `signature` is used instead.
  """
  if new_signature is None:
    new_signature = signature

  # We create a module_fn that creates the new TFHub module.
  def module_fn():
    module = hub.Module(module_path)
    inputs = _placeholders_from_module(module, signature=signature)
    intermediate_tensor = module(inputs, signature=signature, as_dict=True)
    # We need to scope the variables that are created when the transform_fn is
    # applied.
    with tf.variable_scope("transform"):
      outputs = transform_fn(**intermediate_tensor)
    hub.add_signature(name=new_signature, inputs=inputs, outputs=outputs)

  # We create a new graph where we will build the module for export.
  with tf.Graph().as_default():
    # Create the module_spec for the export.
    spec = hub.create_module_spec(module_fn)
    m = hub.Module(spec, trainable=True)
    # We need to recover the scoped variables and remove the scope when loading
    # from the checkpoint.
    prefix = "transform/"
    transform_variables = {
        k[len(prefix):]: v
        for k, v in m.variable_map.items()
        if k.startswith(prefix)
    }
    if transform_variables:
      init_fn = contrib_framework.assign_from_checkpoint_fn(
          transform_checkpoint_path, transform_variables)

    with tf.Session() as sess:
      # Initialize all variables, this also loads the TFHub parameters.
      sess.run(tf.global_variables_initializer())
      # Load the transformer variables from the checkpoint.
      if transform_variables:
        init_fn(sess)
      # Export the new TFHub module.
      m.export(export_path, sess)


def save_numpy_arrays_to_checkpoint(checkpoint_path, **dict_with_arrays):
  """Saves several NumpyArrays to variables in a TF checkpoint.

  Args:
    checkpoint_path: String with the path to the checkpoint file.
    **dict_with_arrays: Dictionary with keys that signify variable names and
      values that are the corresponding Numpy arrays to be saved.
  """
  with tf.Graph().as_default():
    feed_dict = {}
    assign_ops = []
    nodes_to_save = []
    for array_name, array in dict_with_arrays.items():
      # We will save the numpy array with the corresponding dtype.
      tf_dtype = tf.as_dtype(array.dtype)
      # We create a variable which we would like to persist in the checkpoint.
      node = tf.get_variable(array_name, shape=array.shape, dtype=tf_dtype)
      nodes_to_save.append(node)
      # We feed the numpy arrays into the graph via placeholder which avoids
      # adding the numpy arrays to the graph as constants.
      placeholder = tf.placeholder(tf_dtype, shape=array.shape)
      feed_dict[placeholder] = array
      # We use the placeholder to assign the variable the intended value.
      assign_ops.append(tf.assign(node, placeholder))
    saver = tf.train.Saver(nodes_to_save)
    with tf.Session() as sess:
      sess.run(tf.global_variables_initializer())
      sess.run(assign_ops, feed_dict=feed_dict)
      saver.save(sess, checkpoint_path)
  assert saver.last_checkpoints[0] == checkpoint_path


def _placeholders_from_module(tfhub_module, signature):
  """Returns a dictionary with placeholder nodes for a given TFHub module."""
  info_dict = tfhub_module.get_input_info_dict(signature=signature)
  result = {}
  for key, value in info_dict.items():
    result[key] = tf.placeholder(value.dtype, shape=value.get_shape(), name=key)
  return result



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Different functions to extract representations from a Gaussian encoder.

Currently, only the mean of the Gaussian and a random sample from the Gaussian
are supported. However, the interface is set up such that data dependent and
potentially random transformations with learned variables are supported.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import tensorflow.compat.v1 as tf
import gin.tf


@gin.configurable(
    "mean_representation",
    blacklist=["ground_truth_data", "gaussian_encoder", "random_state"])
def mean_representation(
    ground_truth_data,
    gaussian_encoder,
    random_state,
    save_path,
):
  """Extracts the mean representation from a Gaussian encoder.

  Args:
    ground_truth_data: GroundTruthData to be sampled from.
    gaussian_encoder: Function that takes observations as input and outputs a
      dictionary with mean and log variances of the encodings in the keys "mean"
      and "logvar" respectively.
    random_state: Numpy random state used for randomness.
    save_path: String with path where results can be saved.

  Returns:
    transform_fn: Function that takes as keyword arguments the "mean" and
      "logvar" tensors and returns a tensor with the representation.
    None as no variables are saved.

  """
  del ground_truth_data, gaussian_encoder, random_state, save_path

  def transform_fn(mean, logvar):
    del logvar
    return mean

  return transform_fn, None


@gin.configurable(
    "sampled_representation",
    blacklist=["ground_truth_data", "gaussian_encoder", "random_state"])
def sampled_representation(ground_truth_data, gaussian_encoder, random_state,
                           save_path):
  """Extracts the random representation from a Gaussian encoder.

  Args:
    ground_truth_data: GroundTruthData to be sampled from.
    gaussian_encoder: Function that takes observations as input and outputs a
      dictionary with mean and log variances of the encodings in the keys "mean"
      and "logvar" respectively.
    random_state: Numpy random state used for randomness.
    save_path: String with path where results can be saved.

  Returns:
    transform_fn: Function that takes as keyword arguments the "mean" and
      "logvar" tensors and returns a tensor with the representation.
    None as no variables are saved.
  """
  del ground_truth_data, gaussian_encoder, random_state, save_path

  def transform_fn(mean, logvar):
    return tf.add(mean,
                  tf.exp(logvar / 2) * tf.random_normal(tf.shape(mean), 0, 1))

  return transform_fn, None



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Fairness metric task.

In this task, we compute the unfairness of a classifier trained on a given
representation. This procedure is described in Section 4 of the paper "On the
Fairness of Disentangled Representations", Francesco Locatello, Gabriele Abbati,
Tom Rainforth, Stefan Bauer, Bernhard Schoelkopf, Olivier Bachem, NeurIPS 2019,
https://arxiv.org/abs/1905.13662.
The unfairness of a classifier is computed as the total variation distance
between the distribution of predictions of the target varible for different
interventions on the sensitive variable. This approach is described in Section
4 of the paper.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from disentanglement_lib.evaluation.metrics import utils
import numpy as np
from six.moves import range
import gin.tf


@gin.configurable(
    "fairness",
    blacklist=["ground_truth_data", "representation_function", "random_state",
               "artifact_dir"])
def compute_fairness(ground_truth_data,
                     representation_function,
                     random_state,
                     artifact_dir=None,
                     num_train=gin.REQUIRED,
                     num_test_points_per_class=gin.REQUIRED,
                     batch_size=16):
  """Computes unfairness scores.

  We first compute either the mean or maximum total variation for a given
  sensitive and target variable. Then, we either average or take the maximum
  with respect to target and sensitive variable. For convenience, we compute and
  save all combinations. The score used in Section 4 of the paper is here called
  mean_fairness:mean_pred:mean_sens.

  Args:
    ground_truth_data: GroundTruthData to be sampled from.
    representation_function: Function that takes observations as input and
      outputs a dim_representation sized representation for each observation.
    random_state: Numpy random state used for randomness.
    artifact_dir: Optional path to directory where artifacts can be saved.
    num_train: Number of points used for training.
    num_test_points_per_class: Number of points used for testing.
    batch_size: Batch size for sampling.

  Returns:
    Dictionary with scores.
  """
  del artifact_dir
  factor_counts = ground_truth_data.factors_num_values
  num_factors = len(factor_counts)

  scores = {}
  # Training a predictive model.
  mus_train, ys_train = utils.generate_batch_factor_code(
      ground_truth_data, representation_function, num_train, random_state,
      batch_size)
  predictor_model_fn = utils.make_predictor_fn()

  # For each factor train a single predictive model.
  mean_fairness = np.zeros((num_factors, num_factors), dtype=np.float64)
  max_fairness = np.zeros((num_factors, num_factors), dtype=np.float64)
  for i in range(num_factors):
    model = predictor_model_fn()
    model.fit(np.transpose(mus_train), ys_train[i, :])

    for j in range(num_factors):
      if i == j:
        continue
      # Sample a random set of factors once.
      original_factors = ground_truth_data.sample_factors(
          num_test_points_per_class, random_state)
      counts = np.zeros((factor_counts[i], factor_counts[j]), dtype=np.int64)
      for c in range(factor_counts[j]):
        # Intervene on the sensitive attribute.
        intervened_factors = np.copy(original_factors)
        intervened_factors[:, j] = c
        # Obtain the batched observations.
        observations = ground_truth_data.sample_observations_from_factors(
            intervened_factors, random_state)
        representations = utils.obtain_representation(observations,
                                                      representation_function,
                                                      batch_size)
        # Get the predictions.
        predictions = model.predict(np.transpose(representations))
        # Update the counts.
        counts[:, c] = np.bincount(predictions, minlength=factor_counts[i])
      mean_fairness[i, j], max_fairness[i, j] = inter_group_fairness(counts)

  # Report the scores.
  scores.update(compute_scores_dict(mean_fairness, "mean_fairness"))
  scores.update(compute_scores_dict(max_fairness, "max_fairness"))
  return scores


def compute_scores_dict(metric, prefix):
  """Computes scores for combinations of predictive and sensitive factors.

  Either average or take the maximum with respect to target and sensitive
  variable for all combinations of predictive and sensitive factors.

  Args:
    metric: Matrix of shape [num_factors, num_factors] with fairness scores.
    prefix: Prefix for the matrix in the returned dictionary.

  Returns:
    Dictionary containing all combinations of predictive and sensitive factors.
  """
  result = {}
  # Report min and max scores for each predictive and sensitive factor.
  for i in range(metric.shape[0]):
    for j in range(metric.shape[1]):
      if i != j:
        result["{}:pred{}:sens{}".format(prefix, i, j)] = metric[i, j]

  # Compute mean and max values across rows.
  rows_means = []
  rows_maxs = []
  for i in range(metric.shape[0]):
    relevant_scores = [metric[i, j] for j in range(metric.shape[1]) if i != j]
    mean_score = np.mean(relevant_scores)
    max_score = np.amax(relevant_scores)
    result["{}:pred{}:mean_sens".format(prefix, i)] = mean_score
    result["{}:pred{}:max_sens".format(prefix, i)] = max_score
    rows_means.append(mean_score)
    rows_maxs.append(max_score)

  # Compute mean and max values across rows.
  column_means = []
  column_maxs = []
  for j in range(metric.shape[1]):
    relevant_scores = [metric[i, j] for i in range(metric.shape[0]) if i != j]
    mean_score = np.mean(relevant_scores)
    max_score = np.amax(relevant_scores)
    result["{}:sens{}:mean_pred".format(prefix, j)] = mean_score
    result["{}:sens{}:max_pred".format(prefix, j)] = max_score
    column_means.append(mean_score)
    column_maxs.append(max_score)

  # Compute all combinations of scores.
  result["{}:mean_sens:mean_pred".format(prefix)] = np.mean(column_means)
  result["{}:mean_sens:max_pred".format(prefix)] = np.mean(column_maxs)
  result["{}:max_sens:mean_pred".format(prefix)] = np.amax(column_means)
  result["{}:max_sens:max_pred".format(prefix)] = np.amax(column_maxs)
  result["{}:mean_pred:mean_sens".format(prefix)] = np.mean(rows_means)
  result["{}:mean_pred:max_sens".format(prefix)] = np.mean(rows_maxs)
  result["{}:max_pred:mean_sens".format(prefix)] = np.amax(rows_means)
  result["{}:max_pred:max_sens".format(prefix)] = np.amax(rows_maxs)

  return result


def inter_group_fairness(counts):
  """Computes the inter group fairness for predictions based on the TV distance.

  Args:
   counts: Numpy array with counts of predictions where rows correspond to
     predicted classes and columns to sensitive classes.

  Returns:
    Mean and maximum total variation distance of a sensitive class to the
      global average.
  """
  # Compute the distribution of predictions across all sensitive classes.
  overall_distribution = np.sum(counts, axis=1, dtype=np.float32)
  overall_distribution /= overall_distribution.sum()

  # Compute the distribution for each sensitive class.
  normalized_counts = np.array(counts, dtype=np.float32)
  counts_per_class = np.sum(counts, axis=0)
  normalized_counts /= np.expand_dims(counts_per_class, 0)

  # Compute the differences and sum up for each sensitive class.
  differences = normalized_counts - np.expand_dims(overall_distribution, 1)
  total_variation_distances = np.sum(np.abs(differences), 0) / 2.

  mean = (total_variation_distances * counts_per_class)
  mean /= counts_per_class.sum()

  return np.sum(mean), np.amax(total_variation_distances)



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility to create hyperparameter configurations."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import six
from six.moves import range
from six.moves import zip


def sweep(name, values):
  """Sweeps the hyperparameter across different values."""
  return [{name: value} for value in values]


def categorical(items):
  """Defines sweep over categorical variable."""
  return items


def discrete(items):
  """Sweeps over discrete variable."""
  return items


def fixed(name, value, length=1):
  """Creates fixed hyperparameter setting."""
  return [{name: value} for _ in range(length)]


def zipit(list_of_items):
  """Zips different hyperparameter settings."""
  if len(list_of_items) == 1:
    return list_of_items[0]
  main_items = list_of_items[0]
  other_items = zipit(list_of_items[1:])
  if len(main_items) != len(other_items):
    if len(main_items) == 1:
      main_items *= len(other_items)
    elif len(other_items) == 1:
      other_items *= len(main_items)
    else:
      raise ValueError("Cannot zip lists of different lengths.")

  result = []
  for main_dict, other_dict in zip(main_items, other_items):
    new_dict = {}
    new_dict.update(main_dict)
    new_dict.update(other_dict)
    result.append(new_dict)
  return result


def chainit(list_of_items):
  """Chains different hyperparameter settings."""
  result = []
  for items in list_of_items:
    result.extend(items)
  return result


def product(list_of_items):
  """Creates outer product of hyperparameter settings."""
  if len(list_of_items) == 1:
    return list_of_items[0]
  result = []
  other_items = product(list_of_items[1:])
  for first_dict in list_of_items[0]:
    for second_dict in other_items:
      new_dict = {}
      new_dict.update(first_dict)
      new_dict.update(second_dict)
      result.append(new_dict)
  return result


def to_bindings(items):
  return [
      "{} = {}".format(key, _escape_value(value))
      for key, value in items.items()
  ]


def _escape_value(value):
  if isinstance(value, (str, six.text_type)) and not value.startswith("@"):
    return "'{}'".format(value)
  return str(value)



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for util.py."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
# Dependency imports
from absl.testing import parameterized
from disentanglement_lib.data.ground_truth import dummy_data
from disentanglement_lib.data.ground_truth import util
import numpy as np
from six.moves import range
import tensorflow.compat.v1 as tf


class UtilTest(parameterized.TestCase, tf.test.TestCase):

  def test_tfdata(self):
    ground_truth_data = dummy_data.DummyData()
    dataset = util.tf_data_set_from_ground_truth_data(ground_truth_data, 0)
    one_shot_iterator = dataset.make_one_shot_iterator()
    next_element = one_shot_iterator.get_next()
    with self.test_session() as sess:
      for _ in range(10):
        sess.run(next_element)


class StateSpaceAtomIndexTest(parameterized.TestCase, tf.test.TestCase):

  def test(self):
    features = np.array([[0, 1], [1, 0], [1, 1], [0, 0]], dtype=np.int64)
    state_space_atom_index = util.StateSpaceAtomIndex([2, 2], features)
    self.assertAllEqual(
        state_space_atom_index.features_to_index(features), list(range(4)))


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Library of commonly used losses."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy as np
import tensorflow.compat.v1 as tf
import tensorflow_probability as tfp

import gin.tf


@gin.configurable("bernoulli_loss", whitelist=["subtract_true_image_entropy"])
def bernoulli_loss(true_images,
                   reconstructed_images,
                   activation,
                   subtract_true_image_entropy=False):
  """Computes the Bernoulli loss."""
  flattened_dim = np.prod(true_images.get_shape().as_list()[1:])
  reconstructed_images = tf.reshape(
      reconstructed_images, shape=[-1, flattened_dim])
  true_images = tf.reshape(true_images, shape=[-1, flattened_dim])

  # Because true images are not binary, the lower bound in the xent is not zero:
  # the lower bound in the xent is the entropy of the true images.
  if subtract_true_image_entropy:
    dist = tfp.distributions.Bernoulli(
        probs=tf.clip_by_value(true_images, 1e-6, 1 - 1e-6))
    loss_lower_bound = tf.reduce_sum(dist.entropy(), axis=1)
  else:
    loss_lower_bound = 0

  if activation == "logits":
    loss = tf.reduce_sum(
        tf.nn.sigmoid_cross_entropy_with_logits(
            logits=reconstructed_images, labels=true_images),
        axis=1)
  elif activation == "tanh":
    reconstructed_images = tf.clip_by_value(
        tf.nn.tanh(reconstructed_images) / 2 + 0.5, 1e-6, 1 - 1e-6)
    loss = -tf.reduce_sum(
        true_images * tf.log(reconstructed_images) +
        (1 - true_images) * tf.log(1 - reconstructed_images),
        axis=1)
  else:
    raise NotImplementedError("Activation not supported.")

  return loss - loss_lower_bound


@gin.configurable("l2_loss", whitelist=[])
def l2_loss(true_images, reconstructed_images, activation):
  """Computes the l2 loss."""
  if activation == "logits":
    return tf.reduce_sum(
        tf.square(true_images - tf.nn.sigmoid(reconstructed_images)), [1, 2, 3])
  elif activation == "tanh":
    reconstructed_images = tf.nn.tanh(reconstructed_images) / 2 + 0.5
    return tf.reduce_sum(
        tf.square(true_images - reconstructed_images), [1, 2, 3])
  else:
    raise NotImplementedError("Activation not supported.")


@gin.configurable(
    "reconstruction_loss", blacklist=["true_images", "reconstructed_images"])
def make_reconstruction_loss(true_images,
                             reconstructed_images,
                             loss_fn=gin.REQUIRED,
                             activation="logits"):
  """Wrapper that creates reconstruction loss."""
  with tf.variable_scope("reconstruction_loss"):
    per_sample_loss = loss_fn(true_images, reconstructed_images, activation)
  return per_sample_loss



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for visualize_model.py."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from absl.testing import absltest
from absl.testing import parameterized
from disentanglement_lib.methods.unsupervised import train
from disentanglement_lib.utils import resources
from disentanglement_lib.visualize import visualize_model

MODEL_CONFIG_PATH = "../config/tests/methods/unsupervised/train_test.gin"


class VisualizeTest(parameterized.TestCase):

  @parameterized.parameters(
      ("logits"),
      ("tanh"),
  )
  def test_visualize_sigmoid(self, activation):
    activation_binding = (
        "reconstruction_loss.activation = '{}'".format(activation))
    self.model_dir = self.create_tempdir(
        "model_{}".format(activation),
        cleanup=absltest.TempFileCleanup.OFF).full_path
    train.train_with_gin(self.model_dir, True, [
        resources.get_file("config/tests/methods/unsupervised/train_test.gin")
    ], [activation_binding])
    visualize_model.visualize(
        self.model_dir,
        self.create_tempdir("visualization_{}".format(activation)).full_path,
        True,
        num_animations=1,
        num_frames=4,
        num_points_irs=100)


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Library of commonly used architectures and reconstruction losses."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy as np
import tensorflow.compat.v1 as tf
import gin.tf


@gin.configurable("encoder", whitelist=["num_latent", "encoder_fn"])
def make_gaussian_encoder(input_tensor,
                          is_training=True,
                          num_latent=gin.REQUIRED,
                          encoder_fn=gin.REQUIRED):
  """Gin wrapper to create and apply a Gaussian encoder configurable with gin.

  This is a separate function so that several different models (such as
  BetaVAE and FactorVAE) can call this function while the gin binding always
  stays 'encoder.(...)'. This makes it easier to configure models and parse
  the results files.

  Args:
    input_tensor: Tensor with image that should be encoded.
    is_training: Boolean that indicates whether we are training (usually
      required for batch normalization).
    num_latent: Integer with dimensionality of latent space.
    encoder_fn: Function that that takes the arguments (input_tensor,
      num_latent, is_training) and returns the tuple (means, log_vars) with the
      encoder means and log variances.

  Returns:
    Tuple (means, log_vars) with the encoder means and log variances.
  """
  with tf.variable_scope("encoder"):
    return encoder_fn(
        input_tensor=input_tensor,
        num_latent=num_latent,
        is_training=is_training)


@gin.configurable("decoder", whitelist=["decoder_fn"])
def make_decoder(latent_tensor,
                 output_shape,
                 is_training=True,
                 decoder_fn=gin.REQUIRED):
  """Gin wrapper to create and apply a decoder configurable with gin.

  This is a separate function so that several different models (such as
  BetaVAE and FactorVAE) can call this function while the gin binding always
  stays 'decoder.(...)'. This makes it easier to configure models and parse
  the results files.

  Args:
    latent_tensor: Tensor latent space embeddings to decode from.
    output_shape: Tuple with the output shape of the observations to be
      generated.
    is_training: Boolean that indicates whether we are training (usually
      required for batch normalization).
    decoder_fn: Function that that takes the arguments (input_tensor,
      output_shape, is_training) and returns the decoded observations.

  Returns:
    Tensor of decoded observations.
  """
  with tf.variable_scope("decoder"):
    return decoder_fn(
        latent_tensor=latent_tensor,
        output_shape=output_shape,
        is_training=is_training)


@gin.configurable("discriminator", whitelist=["discriminator_fn"])
def make_discriminator(input_tensor,
                       is_training=False,
                       discriminator_fn=gin.REQUIRED):
  """Gin wrapper to create and apply a discriminator configurable with gin.

  This is a separate function so that several different models (such as
  FactorVAE) can potentially call this function while the gin binding always
  stays 'discriminator.(...)'. This makes it easier to configure models and
  parse the results files.

  Args:
    input_tensor: Tensor on which the discriminator operates.
    is_training: Boolean that indicates whether we are training (usually
      required for batch normalization).
    discriminator_fn: Function that that takes the arguments
    (input_tensor, is_training) and returns tuple of (logits, clipped_probs).

  Returns:
    Tuple of (logits, clipped_probs) tensors.
  """
  with tf.variable_scope("discriminator"):
    logits, probs = discriminator_fn(input_tensor, is_training=is_training)
    clipped = tf.clip_by_value(probs, 1e-6, 1 - 1e-6)
  return logits, clipped


@gin.configurable("fc_encoder", whitelist=[])
def fc_encoder(input_tensor, num_latent, is_training=True):
  """Fully connected encoder used in beta-VAE paper for the dSprites data.

  Based on row 1 of Table 1 on page 13 of "beta-VAE: Learning Basic Visual
  Concepts with a Constrained Variational Framework"
  (https://openreview.net/forum?id=Sy2fzU9gl).

  Args:
    input_tensor: Input tensor of shape (batch_size, 64, 64, num_channels) to
      build encoder on.
    num_latent: Number of latent variables to output.
    is_training: Whether or not the graph is built for training (UNUSED).

  Returns:
    means: Output tensor of shape (batch_size, num_latent) with latent variable
      means.
    log_var: Output tensor of shape (batch_size, num_latent) with latent
      variable log variances.
  """
  del is_training

  flattened = tf.layers.flatten(input_tensor)
  e1 = tf.layers.dense(flattened, 1200, activation=tf.nn.relu, name="e1")
  e2 = tf.layers.dense(e1, 1200, activation=tf.nn.relu, name="e2")
  means = tf.layers.dense(e2, num_latent, activation=None)
  log_var = tf.layers.dense(e2, num_latent, activation=None)
  return means, log_var


@gin.configurable("conv_encoder", whitelist=[])
def conv_encoder(input_tensor, num_latent, is_training=True):
  """Convolutional encoder used in beta-VAE paper for the chairs data.

  Based on row 3 of Table 1 on page 13 of "beta-VAE: Learning Basic Visual
  Concepts with a Constrained Variational Framework"
  (https://openreview.net/forum?id=Sy2fzU9gl)

  Args:
    input_tensor: Input tensor of shape (batch_size, 64, 64, num_channels) to
      build encoder on.
    num_latent: Number of latent variables to output.
    is_training: Whether or not the graph is built for training (UNUSED).

  Returns:
    means: Output tensor of shape (batch_size, num_latent) with latent variable
      means.
    log_var: Output tensor of shape (batch_size, num_latent) with latent
      variable log variances.
  """
  del is_training

  e1 = tf.layers.conv2d(
      inputs=input_tensor,
      filters=32,
      kernel_size=4,
      strides=2,
      activation=tf.nn.relu,
      padding="same",
      name="e1",
  )
  e2 = tf.layers.conv2d(
      inputs=e1,
      filters=32,
      kernel_size=4,
      strides=2,
      activation=tf.nn.relu,
      padding="same",
      name="e2",
  )
  e3 = tf.layers.conv2d(
      inputs=e2,
      filters=64,
      kernel_size=2,
      strides=2,
      activation=tf.nn.relu,
      padding="same",
      name="e3",
  )
  e4 = tf.layers.conv2d(
      inputs=e3,
      filters=64,
      kernel_size=2,
      strides=2,
      activation=tf.nn.relu,
      padding="same",
      name="e4",
  )
  flat_e4 = tf.layers.flatten(e4)
  e5 = tf.layers.dense(flat_e4, 256, activation=tf.nn.relu, name="e5")
  means = tf.layers.dense(e5, num_latent, activation=None, name="means")
  log_var = tf.layers.dense(e5, num_latent, activation=None, name="log_var")
  return means, log_var


@gin.configurable("fc_decoder", whitelist=[])
def fc_decoder(latent_tensor, output_shape, is_training=True):
  """Fully connected encoder used in beta-VAE paper for the dSprites data.

  Based on row 1 of Table 1 on page 13 of "beta-VAE: Learning Basic Visual
  Concepts with a Constrained Variational Framework"
  (https://openreview.net/forum?id=Sy2fzU9gl)

  Args:
    latent_tensor: Input tensor to connect decoder to.
    output_shape: Shape of the data.
    is_training: Whether or not the graph is built for training (UNUSED).

  Returns:
    Output tensor of shape (None, 64, 64, num_channels) with the [0,1] pixel
    intensities.
  """
  del is_training
  d1 = tf.layers.dense(latent_tensor, 1200, activation=tf.nn.tanh)
  d2 = tf.layers.dense(d1, 1200, activation=tf.nn.tanh)
  d3 = tf.layers.dense(d2, 1200, activation=tf.nn.tanh)
  d4 = tf.layers.dense(d3, np.prod(output_shape))
  return tf.reshape(d4, shape=[-1] + output_shape)


@gin.configurable("deconv_decoder", whitelist=[])
def deconv_decoder(latent_tensor, output_shape, is_training=True):
  """Convolutional decoder used in beta-VAE paper for the chairs data.

  Based on row 3 of Table 1 on page 13 of "beta-VAE: Learning Basic Visual
  Concepts with a Constrained Variational Framework"
  (https://openreview.net/forum?id=Sy2fzU9gl)

  Args:
    latent_tensor: Input tensor of shape (batch_size,) to connect decoder to.
    output_shape: Shape of the data.
    is_training: Whether or not the graph is built for training (UNUSED).

  Returns:
    Output tensor of shape (batch_size, 64, 64, num_channels) with the [0,1]
      pixel intensities.
  """
  del is_training
  d1 = tf.layers.dense(latent_tensor, 256, activation=tf.nn.relu)
  d2 = tf.layers.dense(d1, 1024, activation=tf.nn.relu)
  d2_reshaped = tf.reshape(d2, shape=[-1, 4, 4, 64])
  d3 = tf.layers.conv2d_transpose(
      inputs=d2_reshaped,
      filters=64,
      kernel_size=4,
      strides=2,
      activation=tf.nn.relu,
      padding="same",
  )

  d4 = tf.layers.conv2d_transpose(
      inputs=d3,
      filters=32,
      kernel_size=4,
      strides=2,
      activation=tf.nn.relu,
      padding="same",
  )

  d5 = tf.layers.conv2d_transpose(
      inputs=d4,
      filters=32,
      kernel_size=4,
      strides=2,
      activation=tf.nn.relu,
      padding="same",
  )
  d6 = tf.layers.conv2d_transpose(
      inputs=d5,
      filters=output_shape[2],
      kernel_size=4,
      strides=2,
      padding="same",
  )
  return tf.reshape(d6, [-1] + output_shape)


@gin.configurable("fc_discriminator", whitelist=[])
def fc_discriminator(input_tensor, is_training=True):
  """Fully connected discriminator used in FactorVAE paper for all datasets.

  Based on Appendix A page 11 "Disentangling by Factorizing"
  (https://arxiv.org/pdf/1802.05983.pdf)

  Args:
    input_tensor: Input tensor of shape (None, num_latents) to build
      discriminator on.
    is_training: Whether or not the graph is built for training (UNUSED).

  Returns:
    logits: Output tensor of shape (batch_size, 2) with logits from
      discriminator.
    probs: Output tensor of shape (batch_size, 2) with probabilities from
      discriminator.
  """
  del is_training
  flattened = tf.layers.flatten(input_tensor)
  d1 = tf.layers.dense(flattened, 1000, activation=tf.nn.leaky_relu, name="d1")
  d2 = tf.layers.dense(d1, 1000, activation=tf.nn.leaky_relu, name="d2")
  d3 = tf.layers.dense(d2, 1000, activation=tf.nn.leaky_relu, name="d3")
  d4 = tf.layers.dense(d3, 1000, activation=tf.nn.leaky_relu, name="d4")
  d5 = tf.layers.dense(d4, 1000, activation=tf.nn.leaky_relu, name="d5")
  d6 = tf.layers.dense(d5, 1000, activation=tf.nn.leaky_relu, name="d6")
  logits = tf.layers.dense(d6, 2, activation=None, name="logits")
  probs = tf.nn.softmax(logits)
  return logits, probs


@gin.configurable("test_encoder", whitelist=["num_latent"])
def test_encoder(input_tensor, num_latent, is_training):
  """Simple encoder for testing.

  Args:
    input_tensor: Input tensor of shape (batch_size, 64, 64, num_channels) to
      build encoder on.
    num_latent: Number of latent variables to output.
    is_training: Whether or not the graph is built for training (UNUSED).

  Returns:
    means: Output tensor of shape (batch_size, num_latent) with latent variable
      means.
    log_var: Output tensor of shape (batch_size, num_latent) with latent
      variable log variances.
  """
  del is_training
  flattened = tf.layers.flatten(input_tensor)
  means = tf.layers.dense(flattened, num_latent, activation=None, name="e1")
  log_var = tf.layers.dense(flattened, num_latent, activation=None, name="e2")
  return means, log_var


@gin.configurable("test_decoder", whitelist=[])
def test_decoder(latent_tensor, output_shape, is_training=False):
  """Simple decoder for testing.

  Args:
    latent_tensor: Input tensor to connect decoder to.
    output_shape: Output shape.
    is_training: Whether or not the graph is built for training (UNUSED).

  Returns:
    Output tensor of shape (batch_size, 64, 64, num_channels) with the [0,1]
      pixel intensities.
  """
  del is_training
  output = tf.layers.dense(latent_tensor, np.prod(output_shape), name="d1")
  return tf.reshape(output, shape=[-1] + output_shape)



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Data sets of Procedurally Generated Matrices (PGMs).

For a description, pleaser refer to https://arxiv.org/abs/1905.12506.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from disentanglement_lib.data.ground_truth import dsprites
from disentanglement_lib.data.ground_truth import dummy_data
from disentanglement_lib.data.ground_truth import ground_truth_data as gtd
from disentanglement_lib.data.ground_truth import named_data
from disentanglement_lib.data.ground_truth import shapes3d
from disentanglement_lib.evaluation.abstract_reasoning import pgm_utils
from disentanglement_lib.utils import resources
from disentanglement_lib.visualize import visualize_util
import gin
import numpy as np
from PIL import Image
import tensorflow.compat.v1 as tf


@gin.configurable("pgm")
def get_pgm_dataset(pgm_type=gin.REQUIRED):
  """Returns a named PGM data set."""
  ground_truth_data = named_data.get_named_ground_truth_data()

  # Quantization for specific data sets (as described in
  # https://arxiv.org/abs/1905.12506).
  if isinstance(ground_truth_data, dsprites.AbstractDSprites):
    wrapped_data_set = Quantizer(ground_truth_data, [5, 6, 3, 3, 4, 4])
  elif isinstance(ground_truth_data, shapes3d.Shapes3D):
    wrapped_data_set = Quantizer(ground_truth_data, [10, 10, 10, 4, 4, 4])
  elif isinstance(ground_truth_data, dummy_data.DummyData):
    wrapped_data_set = ground_truth_data
  else:
    raise ValueError("Invalid data set.")

  # We support different ways to generate PGMs for each of the data set (e.g.,
  # `easy_1`, `hard_3`, `easy_mixes`). `easy` and `hard` refers to the way the
  # alternative solutions of the PGMs are generated:
  #   - `easy`: Alternative answers are random other solutions that do not
  #             satisfy the constraints in the given PGM.
  #   - `hard`: Alternative answers are unique random modifications of the
  #             correct solution which makes the task substantially harder.
  if pgm_type.startswith("easy"):
    sampling = "easy"
  elif pgm_type.startswith("hard"):
    sampling = "hard"
  else:
    raise ValueError("Invalid sampling strategy.")

  # The suffix determines how many relations there are:
  #   - 1-3: Specifies whether always 1, 2, or 3 relations are constant in each
  #          row.
  #   - `mixed`: With probability 1/3 each, 1, 2, or 3 relations are constant
  #               in each row.
  if pgm_type.endswith("1"):
    relations_dist = [1., 0., 0.]
  elif pgm_type.endswith("2"):
    relations_dist = [0., 1., 0.]
  elif pgm_type.endswith("3"):
    relations_dist = [0., 0., 1.]
  elif pgm_type.endswith("mixed"):
    relations_dist = [1. / 3., 1. / 3., 1. / 3.]
  else:
    raise ValueError("Invalid number of relations.")

  return PGMDataset(
      wrapped_data_set,
      sampling_strategy=sampling,
      relations_dist=relations_dist)


class PGMDataset(object):
  """Class that contains PGM data set based on a GroundTruthData."""

  def __init__(self, ground_truth_data, sampling_strategy, relations_dist):
    """Creates a PGMDataset.

    Args:
      ground_truth_data: GroundTruthData data set used to generate images.
      sampling_strategy: Either `easy` or `hard`. For `easy`, alternative
        answers are random other solutions that do not satisfy the constraints
        in the given PGM. For `hard`, alternative answers are unique random
        modifications of the correct solution which makes the task  harder.
      relations_dist: List with probabilites where the i-th element contains the
        probability that i relations are enforced.
    """
    self.ground_truth_data = ground_truth_data
    self.relations_dist = relations_dist
    self.sampling_strategy = sampling_strategy

  def sample(self, random_state):
    """Returns a random PGMInstance."""

    # Sample the number of relations.
    num_relations = 1 + random_state.choice(
        len(self.relations_dist), p=self.relations_dist)

    # Construct the PGM solution in the space of ground-truth factors.
    pgm = pgm_utils.PGM(
        random_state,
        num_relations,
        self.ground_truth_data.factors_num_values,
    )

    # Sample instances of the images for the solutions and alternative answers.
    solution = []
    for row in pgm.matrix:
      solution.append(
          self.ground_truth_data.sample_observations_from_factors(
              row, random_state))

    alternatives = self.ground_truth_data.sample_observations_from_factors(
        pgm.other_solutions, random_state)

    # Sample the position of the correct answer.
    position = random_state.choice(alternatives.shape[0] + 1)
    # Return the instance.
    return PGMInstance(
        np.array(solution), alternatives, position, pgm.matrix,
        pgm.other_solutions, self.ground_truth_data.factors_num_values)

  def tf_data_set(self, seed):
    """Returns a tf.data.Dataset.

    Args:
      seed: Integer with the random seed used to initialize the data set.

    Returns.
      tf.data.Dataset of the data set.
    """

    def generator():
      # We need to hard code the random seed so that the data set can be reset.
      random_state = np.random.RandomState(seed)
      while True:
        instance = self.sample(random_state)
        yield instance.training_sample()

    # We sample a single example to obtain the actual shapes and dtypes.
    features, _ = self.sample(np.random.RandomState(0)).training_sample()
    features_shapes = {k: v.shape for k, v in features.items()}
    features_types = {k: v.dtype for k, v in features.items()}
    output_shapes = (features_shapes, tf.TensorShape([]))
    output_types = (features_types, tf.int64)

    return tf.data.Dataset.from_generator(
        generator, output_types=output_types, output_shapes=output_shapes)

  def make_input_fn(self, seed, num_batches=None):
    """Creates an input function for the TPU Estimator."""

    def input_fn(params):
      """TPUEstimator compatible input fuction."""
      dataset = self.tf_data_set(seed)
      batch_size = params["batch_size"]
      # We need to drop the remainder as otherwise we lose the batch size in the
      # tensor shape. This has no effect as our data set is infinite.
      dataset = dataset.batch(batch_size, drop_remainder=True)
      if num_batches is not None:
        dataset = dataset.take(num_batches)
      return dataset.make_one_shot_iterator().get_next()

    return input_fn


class PGMInstance(object):
  """Class that holds instance of an image PGM."""

  def __init__(self,
               solution,
               alternatives,
               position,
               solution_factors=None,
               alternatives_factors=None,
               num_factor_values=None):
    """Constructs a PGMInstance.

    Args:
      solution: Numpy array of shape (num_rows, num_cols, width, height,
        channels) with the images of the PGM solution.
      alternatives: Numpy array of shape (num_alternatives, width, height,
        channels) with the images of the alternatives.
      position: Integer with position where solution should be inserted.
      solution_factors: Numpy array of shape (num_rows, num_cols, num_factors)
        with the factors of the PGM solution.
      alternatives_factors: Numpy array of shape (num_alternatives, num_factors)
        with the images of the alternatives.
      num_factor_values: List with the number of values for each factor.
    """
    self.solution = solution
    self.alternatives = alternatives
    self.position = position
    self.solution_factors = solution_factors
    self.alternatives_factors = alternatives_factors
    self.num_factor_values = num_factor_values

  def get_context(self):
    """Returns the context.

    Returns:
      Numpy array of shape (num_rows*num_cols - 1, width, height, channels).
    """
    context = []
    for row in self.solution:
      context += list(row)
    return np.array(context[:-1], dtype=np.float32)

  def get_answers(self):
    """Returns the answers.

    Returns:
      Numpy array of shape (num_alternatives + 1, width, height, channels).
    """
    result = list(self.alternatives)
    result.insert(self.position, self.solution[-1, -1])
    return np.array(result, dtype=np.float32)

  def get_context_factor_values(self):
    """Returns the context ground truth factos as integer values.

    Returns:
      Numpy array of shape (num_rows*num_cols - 1, len(num_factor_values).
    """
    context = []
    for row in self.solution_factors:
      context += list(row)
    return np.array(context[:-1])

  def get_answers_factor_values(self):
    """Returns the answers ground truth factos as integer values.

    Returns:
      Numpy array of shape (num_alternatives + 1, len(num_factor_values).
    """
    result = list(self.alternatives_factors)
    result.insert(self.position, self.solution_factors[-1, -1])
    return np.array(result)

  def range_embed_factors(self, factors):
    """Embeds the factors linearly in [-0.5, 0.5] based on integer values.

    Args:
      factors: Numpy array of shape (:, len(num_factor_values) with factors.

    Returns:
      Numpy array of shape (:, len(num_factor_values) with floats.
    """
    result = np.array(factors, dtype=np.float32)
    max_vals = np.array(self.num_factor_values, dtype=np.float32) - 1.
    result /= np.expand_dims(max_vals, 0)
    return result - .5

  def onehot_embed_factors(self, factors):
    """Embeds the factors as one-hot vectors.

    Args:
      factors: Numpy array of shape (:, len(num_factor_values) with factors.

    Returns:
      Numpy array of shape (:, sum(num_factor_values) with floats.
    """
    result = []
    for i, num in enumerate(self.num_factor_values):
      result.append(onehot(factors[:, i], num))
    return np.array(np.concatenate(result, axis=-1), dtype=np.float32)

  def training_sample(self):
    """Returns a single training example."""
    sample = {}
    sample["context"] = self.get_context()
    sample["answers"] = self.get_answers()
    if self.solution_factors is not None:
      context_factors = self.get_context_factor_values()
      answers_factors = self.get_answers_factor_values()

      sample["context_factor_values"] = self.range_embed_factors(
          context_factors)
      sample["answers_factor_values"] = self.range_embed_factors(
          answers_factors)
      sample["context_factors_onehot"] = self.onehot_embed_factors(
          context_factors)
      sample["answers_factors_onehot"] = self.onehot_embed_factors(
          answers_factors)
    return sample, self.position

  def make_image(self, answer=False, padding_px=8, border_px=4):
    """Creates an image of the PGMInstance."""
    # Create the question side that contains the progression matrix.
    question = np.copy(self.solution)
    if question.shape[-1] == 1:
      question = np.repeat(question, 3, -1)
    if not answer:
      question[-1, -1] = question_mark()

    # Build up the image on the context side.
    rows = []
    for i in range(question.shape[0]):
      row = []
      for j in range(question.shape[1]):
        # Do the border around the image.
        color = np.array([1., 1., 1.])
        if answer and i == (question.shape[0] - 1) and j == (question.shape[1] -
                                                             1):
          color = COLORS["green"]
        row.append(
            visualize_util.pad_around(question[i, j], border_px, value=color))
      rows.append(visualize_util.padded_stack(row, padding_px, axis=1))
    question_image = visualize_util.padded_stack(rows, padding_px)

    separator = np.zeros((question_image.shape[0], 2, question_image.shape[2]))

    # Create the answer side.
    answers = self.get_answers()
    if answers.shape[-1] == 1:
      answers = np.repeat(answers, 3, -1)
    answers_with_border = []
    for i, image in enumerate(answers):
      color = np.array([1., 1., 1.])
      if answer:
        color = COLORS["green"] if i == self.position else COLORS["red"]
      answers_with_border.append(
          visualize_util.pad_around(image, border_px, value=color))

    answer_image = visualize_util.padded_grid(answers_with_border,
                                              question.shape[0], padding_px)
    center_crop = visualize_util.padded_stack(
        [question_image, separator, answer_image], padding_px, axis=1)
    return visualize_util.pad_around(
        visualize_util.add_below(center_crop, padding_px), padding_px)


class Quantizer(gtd.GroundTruthData):
  """Quantizes a GroundTruthData to have a maximal number of factors."""

  def __init__(self, wrapped_ground_truth_data, max_factors):
    """Constructs a Quantizer.

    Args:
      wrapped_ground_truth_data: GroundTruthData that should be quantized.
      max_factors: integer with the maximal number of factors.
    """
    self.wrapped_ground_truth_data = wrapped_ground_truth_data
    self.true_num_factors = wrapped_ground_truth_data.factors_num_values
    self.fake_num_factors = list(np.minimum(self.true_num_factors, max_factors))

  @property
  def num_factors(self):
    return self.wrapped_ground_truth_data.num_factors

  @property
  def factors_num_values(self):
    return self.fake_num_factors

  @property
  def observation_shape(self):
    return self.wrapped_ground_truth_data.observation_shape

  def sample_factors(self, num, random_state):
    """Sample a batch of factors Y."""
    factors = np.zeros(shape=(num, self.num_factors), dtype=np.int64)
    for i in range(self.num_factors):
      factors[:, i] = self._sample_factor(i, num, random_state)
    return factors

  def _sample_factor(self, i, num, random_state):
    return random_state.randint(self.factor_sizes[i], size=num)

  def sample_observations_from_factors(self, factors, random_state):
    """Sample a batch of observations X given a batch of factors Y."""
    translated_factors = np.copy(factors)
    for i in range(self.num_factors):
      if self.true_num_factors[i] != self.fake_num_factors[i]:
        ratio = float(self.true_num_factors[i]) / float(
            self.fake_num_factors[i])
        translated_factors[:, i] = np.floor(factors[:, i] * ratio)
    return self.wrapped_ground_truth_data.sample_observations_from_factors(
        translated_factors, random_state)


COLORS = {
    "blue": np.array([66., 103., 210.]) / 255.,
    "red": np.array([234., 67., 53.]) / 255.,
    "yellow": np.array([251., 188., 4.]) / 255.,
    "green": np.array([52., 168., 83.]) / 255.,
    "grey": np.array([154., 160., 166.]) / 255.,
}

QUESTION_MARK = [None]


def question_mark():
  """Returns an image of the question mark."""
  # Cache the image so it is not always reloaded.
  if QUESTION_MARK[0] is None:
    with tf.gfile.Open(
        resources.get_file("google/abstract_reasoning/data/question_mark.png"),
        "rb") as f:
      QUESTION_MARK[0] = np.array(Image.open(f).convert("RGB")) * 1.0 / 255.
  return QUESTION_MARK[0]


def onehot(indices, num_atoms):
  """Embeds the indices as one hot vectors."""
  return np.eye(num_atoms)[indices]



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""setup.py for disentanglement_lib."""

from setuptools import find_packages
from setuptools import setup

setup(
    name='disentanglement_lib',
    version='1.5',
    description=('Library for research on disentangled representations.'),
    author='DisentanglementLib Authors',
    author_email='no-reply@google.com',
    url='http://github.com/google-research/disentanglement_lib',
    license='Apache 2.0',
    packages=find_packages(),
    include_package_data=True,
    scripts=[
        'bin/dlib_aggregate_results',
        'bin/dlib_reproduce',
        'bin/dlib_reason',
        'bin/dlib_visualize_dataset',
        'bin/dlib_evaluate',
        'bin/dlib_udr',
        'bin/dlib_postprocess',
        'bin/dlib_train',
        'bin/dlib_visualize_dataset',
        'bin/dlib_visualize_model',
        'bin/dlib_tests',
        'bin/dlib_download_data',
        'bin/dlib_reproduce_jmlr',
        'bin/dlib_reproduce_semi_supervised',
        'bin/dlib_reproduce_weakly_supervised',
        'bin/dlib_train_semi_supervised',
        'bin/dlib_train_weakly_supervised',
    ],
    install_requires=[
        'future',
        'imageio',
        'gin-config',
        'scikit-learn',
        'numpy',
        'pandas',
        'simplejson',
        'six',
        'matplotlib>=1.5.2',
        'pillow>=5.0.0',
        'pandas>=0.23.0',
        'scipy>=1.0.0',
        'tensorflow_hub>=0.2',
        'tensorflow_probability==0.7',
        'seaborn',
    ],
    extras_require={
        'tf': ['tensorflow==1.14'],
        'tf_gpu': ['tensorflow-gpu==1.14'],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    keywords='tensorflow machine learning disentanglement learning',
)



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for results.py."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import os
from disentanglement_lib.utils import results
import tensorflow.compat.v1 as tf
import gin.tf


@gin.configurable("test")
def test_fn(value=gin.REQUIRED):
  return value


class ResultsTest(tf.test.TestCase):

  def test_namespaced_dict(self):
    """Tests namespacing functionality."""
    base_dict = {"!": "!!"}
    numbers = {"1": "one"}
    chars = {"a": "A"}
    new_dict = results.namespaced_dict(base_dict, numbers=numbers, chars=chars)
    self.assertDictEqual(new_dict, {
        "!": "!!",
        "numbers.1": "one",
        "chars.a": "A"
    })

  def test_dict_to_txt(self):
    """Tests saving functionality to txt file."""
    output_path = os.path.join(self.get_temp_dir(), "export.csv")
    output_dict = {"1": "one"}
    results.save_dict(output_path, output_dict)

  def test_gin_dict_live(self):
    """Tests namespacing functionality based on live gin config."""
    parameter_name = "test.value"
    gin.bind_parameter(parameter_name, 1)
    _ = test_fn()
    self.assertDictEqual(results.gin_dict(), {parameter_name: "1"})

  def test_gin_dict_dir(self):
    """Tests namespacing functionality based on saved gin config."""
    parameter_name = "test.value"
    gin.bind_parameter(parameter_name, 1)
    _ = test_fn()
    config_path = os.path.join(self.get_temp_dir(), "config.gin")
    with tf.gfile.GFile(config_path, "w") as f:
      f.write(gin.operative_config_str())
      f.close()
    self.assertDictEqual(results.gin_dict(config_path), {parameter_name: "1"})

  def test_aggregate_json_results(self):
    """Tests aggregation functionality."""
    tmp_dir = self.get_temp_dir()
    output_path1 = os.path.join(tmp_dir, "export_one.json")
    output_dict1 = {"1": "one"}
    results.save_dict(output_path1, output_dict1)
    output_path2 = os.path.join(tmp_dir, "export_two.json")
    output_dict2 = {"2": "two"}
    results.save_dict(output_path2, output_dict2)
    result_dict = results.aggregate_json_results(tmp_dir)
    self.assertDictEqual(result_dict, {
        "export_one.1": "one",
        "export_two.2": "two"
    })

if __name__ == "__main__":
  tf.test.main()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Hyperparameter sweeps and configs for stage 1 of "abstract_reasoning_study".

Are Disentangled Representations Helpful for Abstract Visual Reasoning?
Sjoerd van Steenkiste, Francesco Locatello, Juergen Schmidhuber, Olivier Bachem.
NeurIPS, 2019.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from disentanglement_lib.config import study
from disentanglement_lib.utils import resources
import disentanglement_lib.utils.hyperparams as h
from six.moves import range


def get_datasets():
  """Returns all the data sets."""
  return h.sweep(
      "dataset.name",
      h.categorical(["shapes3d", "abstract_dsprites"]))


def get_num_latent(sweep):
  return h.sweep("encoder.num_latent", h.discrete(sweep))


def get_seeds(num):
  """Returns random seeds."""
  return h.sweep("model.random_seed", h.categorical(list(range(num))))


def get_default_models():
  """Our default set of models (6 model * 6 hyperparameters=36 models)."""
  # BetaVAE config.
  model_name = h.fixed("model.name", "beta_vae")
  model_fn = h.fixed("model.model", "@vae()")
  betas = h.sweep("vae.beta", h.discrete([1., 2., 4., 6., 8., 16.]))
  config_beta_vae = h.zipit([model_name, betas, model_fn])

  # AnnealedVAE config.
  model_name = h.fixed("model.name", "annealed_vae")
  model_fn = h.fixed("model.model", "@annealed_vae()")
  iteration_threshold = h.fixed("annealed_vae.iteration_threshold", 100000)
  c = h.sweep("annealed_vae.c_max", h.discrete([5., 10., 25., 50., 75., 100.]))
  gamma = h.fixed("annealed_vae.gamma", 1000)
  config_annealed_beta_vae = h.zipit(
      [model_name, c, iteration_threshold, gamma, model_fn])

  # FactorVAE config.
  model_name = h.fixed("model.name", "factor_vae")
  model_fn = h.fixed("model.model", "@factor_vae()")
  discr_fn = h.fixed("discriminator.discriminator_fn", "@fc_discriminator")

  gammas = h.sweep("factor_vae.gamma",
                   h.discrete([10., 20., 30., 40., 50., 100.]))
  config_factor_vae = h.zipit([model_name, gammas, model_fn, discr_fn])

  # DIP-VAE-I config.
  model_name = h.fixed("model.name", "dip_vae_i")
  model_fn = h.fixed("model.model", "@dip_vae()")
  lambda_od = h.sweep("dip_vae.lambda_od",
                      h.discrete([1., 2., 5., 10., 20., 50.]))
  lambda_d_factor = h.fixed("dip_vae.lambda_d_factor", 10.)
  dip_type = h.fixed("dip_vae.dip_type", "i")
  config_dip_vae_i = h.zipit(
      [model_name, model_fn, lambda_od, lambda_d_factor, dip_type])

  # DIP-VAE-II config.
  model_name = h.fixed("model.name", "dip_vae_ii")
  model_fn = h.fixed("model.model", "@dip_vae()")
  lambda_od = h.sweep("dip_vae.lambda_od",
                      h.discrete([1., 2., 5., 10., 20., 50.]))
  lambda_d_factor = h.fixed("dip_vae.lambda_d_factor", 1.)
  dip_type = h.fixed("dip_vae.dip_type", "ii")
  config_dip_vae_ii = h.zipit(
      [model_name, model_fn, lambda_od, lambda_d_factor, dip_type])

  # BetaTCVAE config.
  model_name = h.fixed("model.name", "beta_tc_vae")
  model_fn = h.fixed("model.model", "@beta_tc_vae()")
  betas = h.sweep("beta_tc_vae.beta", h.discrete([1., 2., 4., 6., 8., 10.]))
  config_beta_tc_vae = h.zipit([model_name, model_fn, betas])
  all_models = h.chainit([
      config_beta_vae, config_factor_vae, config_dip_vae_i, config_dip_vae_ii,
      config_beta_tc_vae, config_annealed_beta_vae
  ])
  return all_models


def get_config():
  """Returns the hyperparameter configs for different experiments."""
  arch_enc = h.fixed("encoder.encoder_fn", "@conv_encoder", length=1)
  arch_dec = h.fixed("decoder.decoder_fn", "@deconv_decoder", length=1)
  architecture = h.zipit([arch_enc, arch_dec])
  return h.product([
      get_datasets(),
      architecture,
      get_default_models(),
      get_seeds(5),
  ])


class AbstractReasoningStudyV1(study.Study):
  """Defines the study for the paper."""

  def get_model_config(self, model_num=0):
    """Returns model bindings and config file."""
    config = get_config()[model_num]
    model_bindings = h.to_bindings(config)
    model_config_file = resources.get_file(
        "config/abstract_reasoning_study_v1/stage1/model_configs/shared.gin")
    return model_bindings, model_config_file

  def get_postprocess_config_files(self):
    """Returns postprocessing config files."""
    return list(
        resources.get_files_in_folder(
            "config/abstract_reasoning_study_v1/stage1/postprocess_configs/"))

  def get_eval_config_files(self):
    """Returns evaluation config files."""
    return list(
        resources.get_files_in_folder(
            "config/abstract_reasoning_study_v1/stage1/metric_configs/"))



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Library of commonly used optimizers."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import tensorflow.compat.v1 as tf
import gin.tf


def make_optimizer(optimizer_fn, learning_rate):
  """Wrapper to create the optimizer with a given learning_rate."""
  if learning_rate is None:
    # Learning rate is specified in the optimizer_fn options, or left to its
    # default value.
    return optimizer_fn()
  else:
    # Learning rate is explicitly specified in vae/discriminator optimizer.
    # If it is callable, we assume it's a LR decay function which needs the
    # current global step.
    if callable(learning_rate):
      learning_rate = learning_rate(global_step=tf.train.get_global_step())

    return optimizer_fn(learning_rate=learning_rate)


@gin.configurable("vae_optimizer")
def make_vae_optimizer(optimizer_fn=gin.REQUIRED, learning_rate=None):
  """Wrapper that uses gin to construct an optimizer for VAEs."""
  return make_optimizer(optimizer_fn, learning_rate)


@gin.configurable("discriminator_optimizer")
def make_discriminator_optimizer(optimizer_fn=gin.REQUIRED, learning_rate=None):
  """Wrapper that uses gin to construct an optimizer for the discriminator."""
  return make_optimizer(optimizer_fn, learning_rate)



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Abstract base class for a study."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


class Study(object):
  """Abstract base class used for different studies."""

  def get_model_config(self, model_num=0):
    """Returns model bindings and config file."""
    raise NotImplementedError()

  def print_model_config(self, model_num=0):
    """Prints model bindings and config file."""
    model_bindings, model_config_file = self.get_model_config(model_num)
    print("Gin base config for model training:")
    print("--")
    print(model_config_file)
    print()
    print("Gin bindings for model training:")
    print("--")
    for binding in model_bindings:
      print(binding)

  def get_postprocess_config_files(self):
    """Returns postprocessing config files."""
    raise NotImplementedError()

  def print_postprocess_config(self):
    """Prints postprocessing config files."""
    print("Gin config files for postprocessing (random seeds may be set "
          "later):")
    print("--")
    for path in self.get_postprocess_config_files():
      print(path)

  def get_eval_config_files(self):
    """Returns evaluation config files."""
    raise NotImplementedError()

  def print_eval_config(self):
    """Prints evaluation config files."""
    print("Gin config files for evaluation (random seeds may be set later):")
    print("--")
    for path in self.get_eval_config_files():
      print(path)



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines a common interface for Gaussian encoder based models."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import tensorflow.compat.v1 as tf
import tensorflow_hub as hub
import gin.tf


class GaussianEncoderModel(object):
  """Abstract base class of a Gaussian encoder model."""

  def model_fn(self, features, labels, mode, params):
    """TPUEstimator compatible model function used for training/evaluation."""
    raise NotImplementedError()

  def gaussian_encoder(self, input_tensor, is_training):
    """Applies the Gaussian encoder to images.

    Args:
      input_tensor: Tensor with the observations to be encoded.
      is_training: Boolean indicating whether in training mode.

    Returns:
      Tuple of tensors with the mean and log variance of the Gaussian encoder.
    """
    raise NotImplementedError()

  def decode(self, latent_tensor, observation_shape, is_training):
    """Decodes the latent_tensor to an observation."""
    raise NotImplementedError()

  def sample_from_latent_distribution(self, z_mean, z_logvar):
    """Samples from the Gaussian distribution defined by z_mean and z_logvar."""
    return tf.add(
        z_mean,
        tf.exp(z_logvar / 2) * tf.random_normal(tf.shape(z_mean), 0, 1),
        name="sampled_latent_variable")


@gin.configurable("export_as_tf_hub", whitelist=[])
def export_as_tf_hub(gaussian_encoder_model,
                     observation_shape,
                     checkpoint_path,
                     export_path,
                     drop_collections=None):
  """Exports the provided GaussianEncoderModel as a TFHub module.

  Args:
    gaussian_encoder_model: GaussianEncoderModel to be exported.
    observation_shape: Tuple with the observations shape.
    checkpoint_path: String with path where to load weights from.
    export_path: String with path where to save the TFHub module to.
    drop_collections: List of collections to drop from the graph.
  """

  def module_fn(is_training):
    """Module function used for TFHub export."""
    with tf.variable_scope(tf.get_variable_scope(), reuse=tf.AUTO_REUSE):
      # Add a signature for the Gaussian encoder.
      image_placeholder = tf.placeholder(
          dtype=tf.float32, shape=[None] + observation_shape)
      mean, logvar = gaussian_encoder_model.gaussian_encoder(
          image_placeholder, is_training)
      hub.add_signature(
          name="gaussian_encoder",
          inputs={"images": image_placeholder},
          outputs={
              "mean": mean,
              "logvar": logvar
          })

      # Add a signature for reconstructions.
      latent_vector = gaussian_encoder_model.sample_from_latent_distribution(
          mean, logvar)
      reconstructed_images = gaussian_encoder_model.decode(
          latent_vector, observation_shape, is_training)
      hub.add_signature(
          name="reconstructions",
          inputs={"images": image_placeholder},
          outputs={"images": reconstructed_images})

      # Add a signature for the decoder.
      latent_placeholder = tf.placeholder(
          dtype=tf.float32, shape=[None, mean.get_shape()[1]])
      decoded_images = gaussian_encoder_model.decode(latent_placeholder,
                                                     observation_shape,
                                                     is_training)

      hub.add_signature(
          name="decoder",
          inputs={"latent_vectors": latent_placeholder},
          outputs={"images": decoded_images})

  # Export the module.
  # Two versions of the model are exported:
  #   - one for "test" mode (the default tag)
  #   - one for "training" mode ("is_training" tag)
  # In the case that the encoder/decoder have dropout, or BN layers, these two
  # graphs are different.
  tags_and_args = [
      ({"train"}, {"is_training": True}),
      (set(), {"is_training": False}),
  ]
  spec = hub.create_module_spec(module_fn, tags_and_args=tags_and_args,
                                drop_collections=drop_collections)
  spec.export(export_path, checkpoint_path=checkpoint_path)



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the disentanglement metric from the FactorVAE paper.

Based on "Disentangling by Factorising" (https://arxiv.org/abs/1802.05983).
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from absl import logging
from disentanglement_lib.evaluation.metrics import utils
import numpy as np
from six.moves import range
import gin.tf


@gin.configurable(
    "factor_vae_score",
    blacklist=["ground_truth_data", "representation_function", "random_state",
               "artifact_dir"])
def compute_factor_vae(ground_truth_data,
                       representation_function,
                       random_state,
                       artifact_dir=None,
                       batch_size=gin.REQUIRED,
                       num_train=gin.REQUIRED,
                       num_eval=gin.REQUIRED,
                       num_variance_estimate=gin.REQUIRED):
  """Computes the FactorVAE disentanglement metric.

  Args:
    ground_truth_data: GroundTruthData to be sampled from.
    representation_function: Function that takes observations as input and
      outputs a dim_representation sized representation for each observation.
    random_state: Numpy random state used for randomness.
    artifact_dir: Optional path to directory where artifacts can be saved.
    batch_size: Number of points to be used to compute the training_sample.
    num_train: Number of points used for training.
    num_eval: Number of points used for evaluation.
    num_variance_estimate: Number of points used to estimate global variances.

  Returns:
    Dictionary with scores:
      train_accuracy: Accuracy on training set.
      eval_accuracy: Accuracy on evaluation set.
  """
  del artifact_dir
  logging.info("Computing global variances to standardise.")
  global_variances = _compute_variances(ground_truth_data,
                                        representation_function,
                                        num_variance_estimate, random_state)
  active_dims = _prune_dims(global_variances)
  scores_dict = {}

  if not active_dims.any():
    scores_dict["train_accuracy"] = 0.
    scores_dict["eval_accuracy"] = 0.
    scores_dict["num_active_dims"] = 0
    return scores_dict

  logging.info("Generating training set.")
  training_votes = _generate_training_batch(ground_truth_data,
                                            representation_function, batch_size,
                                            num_train, random_state,
                                            global_variances, active_dims)
  classifier = np.argmax(training_votes, axis=0)
  other_index = np.arange(training_votes.shape[1])

  logging.info("Evaluate training set accuracy.")
  train_accuracy = np.sum(
      training_votes[classifier, other_index]) * 1. / np.sum(training_votes)
  logging.info("Training set accuracy: %.2g", train_accuracy)

  logging.info("Generating evaluation set.")
  eval_votes = _generate_training_batch(ground_truth_data,
                                        representation_function, batch_size,
                                        num_eval, random_state,
                                        global_variances, active_dims)

  logging.info("Evaluate evaluation set accuracy.")
  eval_accuracy = np.sum(eval_votes[classifier,
                                    other_index]) * 1. / np.sum(eval_votes)
  logging.info("Evaluation set accuracy: %.2g", eval_accuracy)
  scores_dict["train_accuracy"] = train_accuracy
  scores_dict["eval_accuracy"] = eval_accuracy
  scores_dict["num_active_dims"] = len(active_dims)
  return scores_dict


@gin.configurable("prune_dims", blacklist=["variances"])
def _prune_dims(variances, threshold=0.):
  """Mask for dimensions collapsed to the prior."""
  scale_z = np.sqrt(variances)
  return scale_z >= threshold


def _compute_variances(ground_truth_data,
                       representation_function,
                       batch_size,
                       random_state,
                       eval_batch_size=64):
  """Computes the variance for each dimension of the representation.

  Args:
    ground_truth_data: GroundTruthData to be sampled from.
    representation_function: Function that takes observation as input and
      outputs a representation.
    batch_size: Number of points to be used to compute the variances.
    random_state: Numpy random state used for randomness.
    eval_batch_size: Batch size used to eval representation.

  Returns:
    Vector with the variance of each dimension.
  """
  observations = ground_truth_data.sample_observations(batch_size, random_state)
  representations = utils.obtain_representation(observations,
                                                representation_function,
                                                eval_batch_size)
  representations = np.transpose(representations)
  assert representations.shape[0] == batch_size
  return np.var(representations, axis=0, ddof=1)


def _generate_training_sample(ground_truth_data, representation_function,
                              batch_size, random_state, global_variances,
                              active_dims):
  """Sample a single training sample based on a mini-batch of ground-truth data.

  Args:
    ground_truth_data: GroundTruthData to be sampled from.
    representation_function: Function that takes observation as input and
      outputs a representation.
    batch_size: Number of points to be used to compute the training_sample.
    random_state: Numpy random state used for randomness.
    global_variances: Numpy vector with variances for all dimensions of
      representation.
    active_dims: Indexes of active dimensions.

  Returns:
    factor_index: Index of factor coordinate to be used.
    argmin: Index of representation coordinate with the least variance.
  """
  # Select random coordinate to keep fixed.
  factor_index = random_state.randint(ground_truth_data.num_factors)
  # Sample two mini batches of latent variables.
  factors = ground_truth_data.sample_factors(batch_size, random_state)
  # Fix the selected factor across mini-batch.
  factors[:, factor_index] = factors[0, factor_index]
  # Obtain the observations.
  observations = ground_truth_data.sample_observations_from_factors(
      factors, random_state)
  representations = representation_function(observations)
  local_variances = np.var(representations, axis=0, ddof=1)
  argmin = np.argmin(local_variances[active_dims] /
                     global_variances[active_dims])
  return factor_index, argmin


def _generate_training_batch(ground_truth_data, representation_function,
                             batch_size, num_points, random_state,
                             global_variances, active_dims):
  """Sample a set of training samples based on a batch of ground-truth data.

  Args:
    ground_truth_data: GroundTruthData to be sampled from.
    representation_function: Function that takes observations as input and
      outputs a dim_representation sized representation for each observation.
    batch_size: Number of points to be used to compute the training_sample.
    num_points: Number of points to be sampled for training set.
    random_state: Numpy random state used for randomness.
    global_variances: Numpy vector with variances for all dimensions of
      representation.
    active_dims: Indexes of active dimensions.

  Returns:
    (num_factors, dim_representation)-sized numpy array with votes.
  """
  votes = np.zeros((ground_truth_data.num_factors, global_variances.shape[0]),
                   dtype=np.int64)
  for _ in range(num_points):
    factor_index, argmin = _generate_training_sample(ground_truth_data,
                                                     representation_function,
                                                     batch_size, random_state,
                                                     global_variances,
                                                     active_dims)
    votes[factor_index, argmin] += 1
  return votes



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Abstract class for data sets that are two-step generative models."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


class GroundTruthData(object):
  """Abstract class for data sets that are two-step generative models."""

  @property
  def num_factors(self):
    raise NotImplementedError()

  @property
  def factors_num_values(self):
    raise NotImplementedError()

  @property
  def observation_shape(self):
    raise NotImplementedError()

  def sample_factors(self, num, random_state):
    """Sample a batch of factors Y."""
    raise NotImplementedError()

  def sample_observations_from_factors(self, factors, random_state):
    """Sample a batch of observations X given a batch of factors Y."""
    raise NotImplementedError()

  def sample(self, num, random_state):
    """Sample a batch of factors Y and observations X."""
    factors = self.sample_factors(num, random_state)
    return factors, self.sample_observations_from_factors(factors, random_state)

  def sample_observations(self, num, random_state):
    """Sample a batch of observations X."""
    return self.sample(num, random_state)[1]
